--[[
  Rise and Shine Incoming Heals for CT_RaidAssist
  Requires: CT_RaidAssist
  
  Written by: Tattersail <Rise and Shine> on Twisting Nether (EU)
  
  IH_IncomingHeals = {
    name = {
      i = { -- if empty, not getting any heals
        hittime = GetTime() + casttime / 1000 - lag
        amount
        source
      } -- sorted by hittime
    }
  }
  IH_OutgoingHeals = {
    targets = { -- if empty, not casting any heal
      i = name
    }
    amount
    hittime
  }

--]]

IH_VERSION = 36
IH_TITLE = "Rise and Shine Incoming Heals"
IH_COLOR_CODE = "00FF00"

IH_AUTHOR_NAME = "Tattersail"
IH_AUTHOR_GUILD = "Rise and Shine"
IH_AUTHOR_REALM = "Twisting Nether EU"

IH_LAG = 25
IH_TIMEOUT = 1 -- timeout heals after GetTime() + (casttime - lag) / 1000 + IH_TIMEOUT
IH_RESTIMEOUT = 20
IH_CHECKHEALS_TIMER = 30
IH_MIN_LOSS_SHOW = 1000 -- players with -HP > X will show IH if ShowMissing is enabled
IH_TOO_LOW_MINUS_HEAL = 0.6 -- heals that hit for less than minimum * X are ignored in +healing calculations (probably target has MS or curse)
IH_SAME_SPELL_DETECT_TIME_RANGE = 0.5 -- combat log msg and calculated hittime are this far apart and spellname is the same --> same spell
IH_SEND_CANCEL_TIME_DIFF = 0.3

IH_IsPaladin = nil

IH_IncomingHeals = { }
IH_OutgoingHeals = { }
IH_OutgoingResses = { }
IH_KnownUsers = { }
IH_GraphFrames = { }
IH_PlayerStats = { }

IH_ShowIncoming = true
IH_ShowOutgoing = false
IH_ShowMissing = false
IH_ShowGraph = true

IH_LastActivatedSpell = nil
IH_CastingSpell = nil
IH_PreviousSpell = nil
IH_OverHealing = nil


function IH_OnVariablesLoaded()
  SLASH_IH1 = "/"..TEXT__IH
  SLASH_IH2 = "/"..TEXT__INCOMINGHEALS
  SlashCmdList["IH"] = IH_SlashHandler

  for i = 1, 40 do
    getglobal("CT_RAMember"..i.."_IH"):SetBackdropColor(0, 0, 0, 1)
    getglobal("CT_RAMember"..i.."_OH"):SetBackdropColor(0, 0, 0, 1)
    getglobal("CT_RAMember"..i.."_OHText"):SetTextColor(0, 1, 0)
  end
  
  for i = 1, 5 do
    getglobal("CT_RA_EmergencyFrameFrame"..i.."_IH"):SetBackdropColor(0, 0, 0, 1)
  end

  for i = 1, 10 do
    getglobal("CT_RAMTGroupMember"..i.."MTTT_IH"):SetBackdropColor(0, 0, 0, 1)
    getglobal("CT_RAPTGroupMember"..i.."_IH"):SetBackdropColor(0, 0, 0, 1)
  end

  hook("CT_RA_ParseMessage", IH_CT_RA_ParseMessage_New, 1)
  hook("CT_RA_AddMessage", IH_CT_RA_AddMessage_New, 200)
  hook("CT_RA_UpdateUnitHealth", IH_CT_RA_UpdateUnitHealth_New, 200)
  hook("CT_RA_UpdateUnitDead", IH_CT_RA_UpdateUnitDead_New, 200)
  hook("CT_RA_Emergency_UpdateHealth", IH_CT_RA_Emergency_UpdateHealth_New, 200)
  hook("CT_RA_UpdateMTTT", IH_CT_RA_UpdateMTTT_New, 200)
  hook("CT_RA_UpdatePT", IH_CT_RA_UpdatePT_New, 200)
  hook("CastSpell", IH_CastSpell_New, 200)
  hook("UseAction", IH_UseAction_New, 1)
  hook("CastSpellByName", IH_CastSpellByName_New, 200)

--  scheduleRepeat("IH_CheckHeals", IH_CHECKHEALS_TIMER, IH_CheckHeals)
  
  local _, class = UnitClass("player")
  IH_IsPaladin = class == "PALADIN"

  IH_TargetGraph:SetMinResize(120, 60)
  if IH_ShowGraph then
    IH_TargetGraph:Show()
    scheduleRepeat("IH_UpdateGraph", 0.001, IH_UpdateGraph)
  else
    IH_TargetGraph:Hide()
  end
  
  schedule(2, echo, "|c00"..IH_COLOR_CODE..IH_TITLE.."|r "..TEXT__VERSION.." "..IH_VERSION.." "..TEXT__INITIALIZED)
end




--[[

  Spell detection part.
  Many thanks to the CT_RaidAssist team.

--]]


function IH_CastSpell_New(spellId, spellbookTabNum)
  IH_LastActivatedSpell = nil
  local oldval = GetCVar("UberTooltips")
  SetCVar("UberTooltips", 1)
  CT_RADST:SetSpell(spellId, spellbookTabNum)
  local spellName, spellRank = GetSpellName(spellId, spellbookTabNum)
  IH_ParseTooltip(spellRank)
  SetCVar("UberTooltips", oldval)
end

function IH_UseAction_New(a1, a2, a3)
  IH_LastActivatedSpell = nil
  local oldval = GetCVar("UberTooltips")
  SetCVar("UberTooltips", 1)
  CT_RADST:SetAction(a1)
  IH_ParseTooltip()
  SetCVar("UberTooltips", oldval)
end

function IH_CastSpellByName_New(spellName)
  IH_LastActivatedSpell = nil
  local _, _, spellname = string.find(spellName, "^([^%(]+) %(");
  local _, _, ranktext = string.find(spellName, "%(([^%(]+)%)");

  if not spellname then spellname = spellName end

  local spelldata = IH_SpellData[spellname]
  if spelldata then
    local rankdata = nil

    if ranktext then
      rankdata = spelldata.ranks[ranktext]

    else
      local maxamount = 0
      for rt, rd in spelldata.ranks do
        if rd.amount and rd.amount > maxamount then
          rankdata = rd
          ranktext = rt
          maxamount = rd.amount
        end
      end
    end

    if not ranktext then
      ranktext = "unknown"
      rankdata = spelldata.ranks[ranktext]
    end

    if not rankdata then
      rankdata = { plushealing = 0, amount = 0, minimum = -200, maximum = 200 }
      spelldata.ranks[ranktext] = rankdata
    end

    IH_LastActivatedSpell = {
      spellname = spellname,
      ranktext = ranktext,
      spelldata = spelldata,
      rankdata = rankdata,
    }
  end
end


function IH_ParseTooltip(ranktext)
  local spellname = CT_RADSTTextLeft1:GetText()
  local spelldata = IH_SpellData[spellname]
  if not spelldata then
    IH_LastActivatedSpell = nil
    return
  end
  
  local minmaxtext = CT_RADSTTextLeft4:GetText()
  if not minmaxtext then minmaxtext = CT_RADSTTextLeft3:GetText() end
  
  local _, _, minimum, maximum = string.find(minmaxtext, IH_TOOLTIP_REGEXP)

  if not ranktext then ranktext = CT_RADSTTextRight1:GetText() end
  if not ranktext then ranktext = minimum end
  local rankdata = spelldata.ranks[ranktext]
  
  if not rankdata then
    rankdata = { plushealing = 0 }
    spelldata.ranks[ranktext] = rankdata
  end

  rankdata.minimum = tonumber(minimum)
  rankdata.maximum = tonumber(maximum)
  rankdata.amount = floor((minimum + maximum) / 2)

  IH_LastActivatedSpell = {
    spellname = spellname,
    ranktext = ranktext,
    spelldata = spelldata,
    rankdata = rankdata,
  }
  
end


function IH_OnSpellCastStart()
  if IH_LastActivatedSpell then
    local spelldata = IH_LastActivatedSpell.spelldata
    if spelldata.party then
      IH_LastActivatedSpell.target = UnitName("player")
      for i = 1, GetNumPartyMembers() do
        IH_LastActivatedSpell.target = IH_LastActivatedSpell.target.." "..UnitName("party"..i)
      end
    elseif CT_RA_SpellCast and CT_RA_SpellCast[2] then -- CTRA spelldetect sets this variable to name of target
      IH_LastActivatedSpell.target = string.gsub(CT_RA_SpellCast[2], " ", "_") -- NPCs can have spaces in their names, players cannot
    end

    if IH_LastActivatedSpell.target then
      local rankdata = IH_LastActivatedSpell.rankdata
      local amount = rankdata.amount + rankdata.plushealing
      if IH_IsPaladin then
        if
          spelldata.bol and
          CT_RA_Stats[IH_LastActivatedSpell.target] and
          CT_RA_Stats[IH_LastActivatedSpell.target]["Buffs"] and
          CT_RA_Stats[IH_LastActivatedSpell.target]["Buffs"][IH_BLESSING_OF_LIGHT]
        then
          amount = amount + spelldata.bol
        end
        for i = 1, 16 do
--          if UnitBuff("player", i) == "Interface\\Icons\\Spell_Holy_WordFortitude" then -- divine favor gives 100% crit chance on next holy spell
          if UnitBuff("player", i) == "Interface\\Icons\\Spell_Holy_Heal" then -- divine favor gives 100% crit chance on next holy spell
            amount = floor(amount * 1.5)
          end
        end
      end
      
--      spelldata.casttime = arg2
      IH_LastActivatedSpell.amount = amount
      IH_LastActivatedSpell.hittime = GetTime() + arg2 / 1000
      IH_PreviousSpell = IH_CastingSpell
      IH_CastingSpell = IH_LastActivatedSpell
      CT_RA_SendMessage("IHN "..amount.." "..arg2.." "..IH_LastActivatedSpell.target, true) -- arg2 is casttime in ms
--      echo("IHN "..amount.." "..arg2.." "..IH_LastActivatedSpell.target)
      IH_LastActivatedSpell = nil
    else
      IH_PreviousSpell = IH_CastingSpell
      IH_CastingSpell = nil
    end
  end
end

function IH_OnSpellCastStop()
  if IH_CastingSpell and IH_CastingSpell.target and IH_CastingSpell.amount and IH_CastingSpell.hittime - GetTime() > IH_SEND_CANCEL_TIME_DIFF then
    CT_RA_SendMessage("IHS "..IH_CastingSpell.amount.." "..IH_CastingSpell.target, true)
    IH_CastingSpell = nil
  end
end

function IH_OnSpellCastDelayed()
  if IH_CastingSpell and IH_CastingSpell.target and IH_CastingSpell.amount then
    CT_RA_SendMessage("IHD "..arg1, true) -- arg1 is delay time in ms
  end
end


function IH_OnMsgSpellSelfBuff() -- spell finishes casting
  local spell = nil
  local _, _, spellname, target, amount = string.find(arg1, IH_PLUSHEALING_REGEXP)
  
  for _, spell2 in { IH_CastingSpell, IH_PreviousSpell } do
    if
      spell2 and
      spell2.spellname == spellname and
      not spell2.plushealfixed and
      spell2.hittime and
      math.abs(spell2.hittime - GetTime()) < IH_SAME_SPELL_DETECT_TIME_RANGE
    then
      spell = spell2
    end
  end

  if spell then
    local rankdata = spell.rankdata
    local spelldata = spell.spelldata

    amount = tonumber(amount)

    if IH_IsPaladin then
      if target == "you" then target = UnitName("player") end
      if
        spell.spelldata.bol and
        CT_RA_Stats[target] and
        CT_RA_Stats[target]["Buffs"] and
        CT_RA_Stats[target]["Buffs"][IH_BLESSING_OF_LIGHT]
      then
        amount = amount - spelldata.bol
      end
    end


    -- calculate + heal
    if amount > rankdata.maximum then rankdata.plushealing = max(rankdata.plushealing, amount - rankdata.maximum) end

    -- calculate - heal (when you remove equipment or it breaks)
    if amount < rankdata.minimum + rankdata.plushealing and amount > (rankdata.minimum + rankdata.plushealing) * IH_TOO_LOW_MINUS_HEAL then
      rankdata.plushealing = min(rankdata.plushealing, amount - rankdata.minimum)
    end

    spell.plushealfixed = true
  end
end


--[[

  Parse incoming CT_RaidAssist and IncomingHeals messages.

  RES <player>
  RESNO
  IHN <amount> <casttime> <player>
  IHH <amount> <duration> <player>
  IHS <amount> <player>
  IHD <time>
  IHV <version>
  
  RES - starts res on <player>
  RESNO - stop ressing
  IHN - normal heal
  IHH - HoT - not implemented yet. impossible atm.
  IHS - stop heal - only used when interrupted. 
  IHD - delayed by amount of time
  
--]]

function IH_CT_RA_ParseMessage_New(source, msg)
  if string.sub(msg, 1, 4) == "RES " then
    local _, _, target = string.find(msg, "^RES (.+)$");
    
    if not IH_OutgoingResses[source] then IH_OutgoingResses[source] = { } end
    local outres = IH_OutgoingResses[source]
    outres.target = target
    outres.time = GetTime()
    
    local raidid = GetPlayerRaidID(target)
    if raidid then
      getglobal("CT_RAMember"..raidid.."_IH"):Show()
      getglobal("CT_RAMember"..raidid.."_IHText"):SetText("RI")
      getglobal("CT_RAMember"..raidid.."_IHText"):SetTextColor(1, 1, 0)
    end

  elseif string.sub(msg, 1, 5) == "RESNO" then
    local outres = IH_OutgoingResses[source]
    if outres and outres.target then
      local target = outres.target
      local moreres = false
      outres.target = nil
      
      for source, outres in IH_OutgoingResses do
        if outres.target and outres.target == target then moreres = true end
      end

      if not moreres then
        local raidid = GetPlayerRaidID(target)
        if raidid then
          getglobal("CT_RAMember"..raidid.."_IH"):Hide()
        end
      end
    end

  elseif string.sub(msg, 1, 4) == "IHN " then
    if not IH_KnownUsers[source] then IH_KnownUsers[source] = 7 end
    IH_FinishHeal(source)

    local _, _, amount, casttime, target = string.find(msg, "^IHN (%d+) (%d+) (.+)$")
    if amount and casttime and target then
    
      local targets = string.split(target, " ")
      if table.getn(targets) > 1 then target = "multiple" end
      for i = 1, table.getn(targets) do
        targets[i] = string.gsub(targets[i], "_", " ")
      end

      if not IH_OutgoingHeals[source] then IH_OutgoingHeals[source] = { } end
      local oh = IH_OutgoingHeals[source]
      oh.targets = targets
      oh.amount = amount
      oh.hittime = GetTime() + (casttime - IH_LAG) / 1000
      
      scheduleByName("IHO_"..source, (casttime - IH_LAG) / 1000, IH_FinishHeal, source)

      if IH_ShowOutgoing then
        local raidid = GetPlayerRaidID(source)
        if raidid then
          getglobal("CT_RAMember"..raidid.."_OH"):Show()
          getglobal("CT_RAMember"..raidid.."_OHText"):SetText("-> "..target)
        end
      end

      for index, target in targets do
        if not IH_IncomingHeals[target] then IH_IncomingHeals[target] = { } end
        local ihs = IH_IncomingHeals[target]
        table.insert(ihs, {
          source = source,
          amount = amount,
          hittime = GetTime() + casttime / 1000 - IH_LAG / 1000
        })
        table.sort(ihs, function(t1, t2) return t1.hittime < t2.hittime end)

        IH_UpdateTarget(target)
      end
      
      if not IH_PlayerStats[source] then IH_PlayerStats[source] = { heals = 0, cancels = 0 } end
      IH_PlayerStats[source].heals = IH_PlayerStats[source].heals + 1

    end
    return
  
  elseif string.sub(msg, 1, 4) == "IHD " then
    local _, _, time = string.find(msg, "^IHD (%d+)$")
    local oh = IH_OutgoingHeals[source]
    if oh and oh.targets then
      oh.hittime = oh.hittime + time / 1000
      TimersByName["IHO_"..source].endtime = oh.hittime
      for index, target in oh.targets do
        local ihs = IH_IncomingHeals[target]
        if ihs and table.getn(ihs) > 0 then
          for index, ih in ihs do
            if ih.source == source then
              ih.hittime = ih.hittime + time / 1000
            end
          end
          table.sort(ihs, function(t1, t2) return t1.hittime < t2.hittime end)
          IH_UpdateTarget(target)
        end
      end
    end
    
  elseif string.sub(msg, 1, 4) == "IHV " then
    local _, _, version = string.find(msg, "^IHV (%d+)$")
    IH_KnownUsers[source] = tonumber(version)
    return
  
  elseif string.sub(msg, 1, 4) == "IHS " then
    local _, _, amount, targets = string.find(msg, "^IHS (%d+) (.+)$")

    if not IH_PlayerStats[source] then IH_PlayerStats[source] = { heals = 0, cancels = 0 } end
    IH_PlayerStats[source].cancels = IH_PlayerStats[source].cancels + 1
    
    IH_FinishHeal(source)
    return
  
  end
end

function IH_FinishHeal(source)
  unscheduleByName("IHO_"..source)
  
  if IH_ShowOutgoing then
    local raidid = GetPlayerRaidID(source)
    if raidid then getglobal("CT_RAMember"..raidid.."_OH"):Hide() end
  end

  local oh = IH_OutgoingHeals[source]
  if oh and oh.targets then
    for index, target in oh.targets do
      local ihs = IH_IncomingHeals[target]
      if ihs and table.getn(ihs) > 0 then
        table.foreach(ihs, function(index, ih) if ih.source == source then table.remove(ihs, index) end end)
      end
      table.sort(ihs, function(t1, t2) return t1.hittime < t2.hittime end)
      IH_UpdateTarget(target)
    end
    oh.targets = nil
  end


end

--[[
  Periodically checks all incoming and outgoing heals.
  Lag and bugs can cause values to become "stuck".
  This check should detect this and remove them.
--]]

--[[
function IH_CheckHeals()
  for source, oh in IH_OutgoingHeals do
    if oh.targets and oh.hittime + IH_TIMEOUT < GetTime() then
      IH_FinishHeal(source)
    end
  end
  
  table.foreach(IH_OutgoingResses, function(source, outres) if outres.target and outres.time + IH_RESTIMEOUT < GetTime() then IH_OutgoingResses[source] = nil end end)
  
  for raidid = 1, GetNumRaidMembers() do
    local name = GetRaidRosterInfo(raidid)
    
    if getglobal("CT_RAMember"..raidid.."_IH"):IsVisible() then
      local hide = true
      if IH_IncomingHeals[name] and IH_IncomingHeals[name].amount then hide = false end
      if UnitHealth("raid"..raidid) < UnitHealthMax("raid"..raidid) - IH_MIN_LOSS_SHOW then hide = false end
      if hide then
        for source, outres in IH_OutgoingResses do
          if outres and outres.target == name then hide = false end
        end
      end
      if not IH_ShowMissing then hide = true end
      if hide then
        getglobal("CT_RAMember"..raidid.."_IH"):Hide()
      end
    end
    
    if getglobal("CT_RAMember"..raidid.."_OH"):IsVisible() then
      local hide = true
      for source, oh in IH_OutgoingHeals do
        if oh and oh.targets and table.getkey(oh.targets, name) then hide = false end
      end
      if hide then
        getglobal("CT_RAMember"..raidid.."_OH"):Hide()
      end
    end
  end
end
--]]


function IH_CT_RA_UpdateUnitHealth_New(frame)
  local raidid = frame:GetID()
  local unitid = "raid"..raidid
  local name = UnitName(unitid)
  IH_UpdateTarget(name, unitid, raidid)
end

function IH_CT_RA_UpdateUnitDead_New(frame, didUpdateHealth)
  local raidid = frame.id
  local unitid = "raid"..raidid
  local name = UnitName(unitid)
  IH_UpdateTarget(name, unitid, raidid)
end


--[[
  Update the IH numbers on CTRA Emergency Monitor
--]]

function IH_CT_RA_Emergency_UpdateHealth_New()
  for i = 1, 5 do
    if getglobal("CT_RA_EmergencyFrameFrame"..i):IsVisible() then
      local name = getglobal("CT_RA_EmergencyFrameFrame"..i.."Name"):GetText()
      IH_UpdateTarget(name)
    end
  end
end





--[[    MTTT display    ]]--

function IH_UpdateMTTTs()
  for i = 1, 10 do
    getglobal("CT_RAMTGroupMember"..i.."MTTT").unitName = nil
  end
  CT_RA_UpdateMTTTs()
end

function IH_CT_RA_UpdateMTTT_New(raidid, mtid, frame, height, key, val)
  local framename = frame:GetName()
  local name = UnitName(raidid)
  IH_UpdateTarget(name, raidid)
end



--[[    PT display    ]]--

function IH_UpdatePTs()
  for i = 1, 10 do
    getglobal("CT_RAPTGroupMember"..i).unitName = nil
  end
  CT_RA_UpdatePTs()
end

function IH_CT_RA_UpdatePT_New(raidid, frame, height, key, val)
  local framename = frame:GetName()
  local name = UnitName(raidid)
  IH_UpdateTarget(name, raidid)
end



function IH_UpdateTarget(target, unitid, raidid)
  if not raidid then raidid = GetPlayerRaidID(target) end
  if not unitid then
    if raidid then
      unitid = "raid"..raidid
    else
      for i = 1, GetNumRaidMembers() do
        if UnitName("raid"..i.."targettarget") == target then unitid = "raid"..i.."targettarget" end
      end
    end
  end
  
  if not unitid then return end
  
  local frames = { }
  if raidid then table.insert(frames, "CT_RAMember"..raidid) end
  for index, _ in CT_RA_MainTanks do
    local framename = "CT_RAMTGroupMember"..index.."MTTT"
    if getglobal(framename).unitName == target then table.insert(frames, framename) end
  end
  for index, name in CT_RA_PTargets do
    if name == target then table.insert(frames, "CT_RAPTGroupMember"..index) end
  end
  for i = 1, 5 do
    if getglobal("CT_RA_EmergencyFrameFrame"..i).Name:GetText() == target then table.insert(frames, "CT_RA_EmergencyFrameFrame"..i) end
  end
  
  local ihs = IH_IncomingHeals[target]
  if (IH_ShowMissing and UnitHealth(unitid) < UnitHealthMax(unitid) - IH_MIN_LOSS_SHOW and not UnitIsDeadOrGhost(unitid)) or ihs and table.getn(ihs) > 0 then
    local amount = 0
    local color = 1
    if ihs and table.getn(ihs) > 0 then
      for i, ih in ihs do
        amount = amount + ih.amount
        if ih.source == UnitName("player") then
          color = 1 - min(max((UnitHealth(unitid) + amount - UnitHealthMax(unitid)) / ih.amount, 0), 1)
          IH_OverHealing = color == 0
        end
      end
    end

    if IH_ShowMissing then amount = amount - UnitHealthMax(unitid) + UnitHealth(unitid) end
    
    for _, framename in frames do
      getglobal(framename.."_IH"):Show()
      getglobal(framename.."_IHText"):SetText(string.format("%.1f", amount / 1000))
      getglobal(framename.."_IHText"):SetTextColor(1, color, color)
    end

  else
    for _, framename in frames do
      getglobal(framename.."_IH"):Hide()
    end

  end
  
  if UnitName("target") == target then IH_UpdateGraph() end
end


function IH_OnTargetChanged()
  IH_UpdateGraph()
end


function IH_UpdateGraph()
  local target = UnitName("target")
  
  if IH_ShowGraph and GetNumRaidMembers() > 0 and target then
    IH_TargetGraph:Show()
    IH_TargetGraph:SetBackdropColor(0, 0, 0, 1)
    IH_TargetGraphMissingNow:SetText(string.format("%.1f", (UnitHealth("target") - UnitHealthMax("target")) / 1000))
    IH_TargetGraphName:SetText(target)
    if IH_TargetGraphName:GetStringWidth() > IH_TargetGraph:GetWidth() - 60 then
      IH_TargetGraphName:Hide()
    else
      local _, class = UnitClass("target")
      local color = RAID_CLASS_COLORS[class]
      IH_TargetGraphName:Show()
      IH_TargetGraphName:SetTextColor(color.r, color.g, color.b)
    end
    for index, frame in IH_GraphFrames do
      frame:Hide()
    end
    local ihs = IH_IncomingHeals[target]
    local amount = 0
    if ihs and table.getn(ihs) > 0 then
      for index, ih in ihs do
        local frame = IH_GraphFrames[index]
        if not frame then
          frame = CreateFrame("Frame", "IH_TargetGraphBar"..index, IH_TargetGraph)
          frame:SetWidth(5)
          frame:SetBackdrop({bgFile = "Interface/ChatFrame/ChatFrameBackground",  tile = false, tileSize = 0, edgeSize = 0,
                               insets = { left = 0, right = 0, top = 0, bottom = 0 }})
          frame:SetBackdropColor(1, 1, 1, 1)
          frame:SetParent(IH_TargetGraph)
          IH_GraphFrames[index] = frame
        end
        frame:ClearAllPoints()
        frame:SetPoint("BOTTOMLEFT", "IH_TargetGraph", "BOTTOMLEFT", floor((ih.hittime - GetTime()) / 3 * IH_TargetGraph:GetWidth()), 0)
        frame:SetHeight(floor(ih.amount / 3000 * IH_TargetGraph:GetHeight()))
        
        amount = amount + ih.amount
        if ih.source == UnitName("player") then
          local overheal = min(max((UnitHealth("target") + amount - UnitHealthMax("target")) / ih.amount, 0), 1)
          --frame:SetBackdropColor(overheal, (1 - overheal) / 2, 1 - overheal, 1)
          frame:SetBackdropColor(0, 0.5, 1, 1)
          IH_TargetGraph:SetBackdropColor(overheal / 4, 0, 0, 1)
        else
          frame:SetBackdropColor(1, 1, 1, 1)
        --  frame:SetBackdropColor(1 - overheal, 1, 1 - overheal, 1)
        end
        frame:Show()
      end
    end
    IH_TargetGraphMissingAfter:SetText(string.format("%.1f", (UnitHealth("target") + amount - UnitHealthMax("target")) / 1000))
  else
    IH_TargetGraph:Hide()
  end
end

function IH_ScaleGraph(scale)
  if IH_ShowGraph then
    IH_TargetGraph:ClearAllPoints()
    IH_TargetGraph:SetPoint("TOPLEFT", "UIParent", "TOPLEFT", 400, -400)
    IH_TargetGraph:SetScale(scale)
  end
end


--[[
  Manual reset of all incoming and outgoing heals.
--]]
function IH_Reset()
  IH_IncomingHeals = { }
  IH_OutgoingHeals = { }
  IH_OutgoingResses = { }
  
  for rank, rankdata in IH_SpellData do
    rankdata.ranks = { }
  end

  for i = 1, 40 do
    getglobal("CT_RAMember"..i.."_OH"):Hide()
    getglobal("CT_RAMember"..i.."_IH"):Hide()
  end
  for i = 1, 5 do
    getglobal("CT_RA_EmergencyFrameFrame"..i.."_IH"):Hide()
  end
  for i = 1, 10 do
    getglobal("CT_RAMTGroupMember"..i.."MTTT_IH"):Hide()
    getglobal("CT_RAPTGroupMember"..i.."_IH"):Hide()
  end

end



--[[
  Sends IHV when CT_RaidAssist sends V
--]]
function IH_CT_RA_AddMessage_New(msg)
  if string.sub(msg, 1, 1) == "V" then
    CT_RA_AddMessage("IHV "..IH_VERSION)
  end
end



function GetPlayerRaidID(playername)
  for i = 1, GetNumRaidMembers(), 1 do
    local name, rank, _, _, _, _, _, _, _ = GetRaidRosterInfo(i)
    if playername == name and rank then
      return i
    end
  end
  return nil
end





function IH_SwitchInDisplay()
  if IH_ShowIncoming then
    echo(TEXT_INCOMING_HEALS_DISPLAY_TURNED.." |c00E02020"..TEXT__OFF.."|r")
    IH_ShowIncoming = false
  else
    echo(TEXT_INCOMING_HEALS_DISPLAY_TURNED.." |c0060F060"..TEXT__ON.."|r")
    IH_ShowIncoming = true
  end
end

function IH_SwitchOutDisplay()
  if IH_ShowOutgoing then
    echo(TEXT_OUTGOING_HEALS_DISPLAY_TURNED.." |c00E02020"..TEXT__OFF.."|r")
    for i = 1, 40 do
      getglobal("CT_RAMember"..i.."_OH"):Hide()
    end
    IH_ShowOutgoing = false
    
  else
    echo(TEXT_OUTGOING_HEALS_DISPLAY_TURNED.." |c0060F060"..TEXT__ON.."|r")
    IH_ShowOutgoing = true

  end
end

function IH_SwitchMissingDisplay()
  if IH_ShowMissing then
    echo(TEXT_MISSING_LIFE_DISPLAY_TURNED.." |c00E02020"..TEXT__OFF.."|r")
    IH_ShowMissing = false
  else
    echo(TEXT_MISSING_LIFE_DISPLAY_TURNED.." |c0060F060"..TEXT__ON.."|r")
    IH_ShowMissing = true
  end
end

function IH_SwitchGraph()
  if IH_ShowGraph then
    IH_TargetGraph:Hide()
    unscheduleByName("IH_UpdateGraph", 0.01, IH_UpdateGraph)
    IH_ShowGraph = false
    echo(TEXT_GRAPH_DISPLAY_TURNED.." |c00E02020"..TEXT__OFF.."|r")
  else
    IH_TargetGraph:Show()
    scheduleRepeat("IH_UpdateGraph", 0.01, IH_UpdateGraph)
    IH_ShowGraph = true
    echo(TEXT_GRAPH_DISPLAY_TURNED.." |c0060F060"..TEXT__ON.."|r")
  end
end


function IH_ShowUsers()
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r - "..TEXT__KNOWN_USERS..":")
  local users = { }
  for playername, version in IH_KnownUsers do
    table.insert(users, { name = playername, version = version })
  end
  table.sort(users, function(t1, t2) return t1.name < t2.name end)
  for index, user in users do
    if user.version <= 7 then user.version = "<= 7" end
    echo(user.name.." "..TEXT__USING_VERSION.." "..user.version)
  end
  
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r - "..TEXT__RAID_HEALERS_NOT_USING_INCOMING_HEALS..":")
  for i = 1, GetNumRaidMembers() do
    local _, class = UnitClass("raid"..i)
    if (class == "PRIEST" or class == "DRUID" or class == "PALADIN" or class == "SHAMAN") and not IH_KnownUsers[UnitName("raid"..i)] then
      echo(UnitName("raid"..i))
    end
  end
end

function IH_ShowStats()
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r "..TEXT__STATISTICS)

  local users = { }
  for playername, stats in IH_PlayerStats do
    table.insert(users, { name = playername, heals = stats.heals, cancels = stats.cancels })
  end
  table.sort(users, function(t1, t2) return t1.heals > t2.heals end)
  for index, user in users do
    echo(user.name.." "..TEXT__HEALS..": "..user.heals.." "..TEXT__CANCELS..": "..user.cancels)
  end
end

function IH_ResetStats()
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r "..TEXT__STATISTICS_RESETTED)
  IH_PlayerStats = { }
end

function IH_SlashHandler(msg)
  local params = string.split(msg, " ")
  if params[1] == TEXT__IN then IH_SwitchInDisplay()
  elseif params[1] == TEXT__MISSING then IH_SwitchMissingDisplay()
  elseif params[1] == TEXT__OUT then IH_SwitchOutDisplay()
--  elseif params[1] == TEXT__GRAPH and params[2] == TEXT__SCALE then IH_ScaleGraph(tonumber(params[3]))
  elseif params[1] == TEXT__GRAPH then IH_SwitchGraph()
  elseif params[1] == TEXT__STATS and params[2] == TEXT__RESET then IH_ResetStats()
  elseif params[1] == TEXT__STATS then IH_ShowStats()
  elseif params[1] == TEXT__USERS then IH_ShowUsers()
  elseif params[1] == TEXT__RESET then IH_Reset()
  elseif params[1] == TEXT__VERSION then IH_ShowVersion()
  elseif msg == "test" then
    echo("Buffs for "..UnitName("target"))
    for i = 1, 16 do
      local buff = UnitBuff("target", i)
      if buff then echo(buff) end
    end
  else IH_ShowHelp()
  end
end

function IH_ShowHelp()
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r "..TEXT__COMMANDS..":")
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__IN.."|r "..TEXT_SWITCHES_DISPLAY_OF_INCOMING_HEALS_ON__S__OFF)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__MISSING.."|r "..TEXT_SWITCHES_DISPLAY_OF_MISSING_LIFE_ON__S__OFF)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__OUT.."|r "..TEXT_SWITCHES_DISPLAY_OF_OUTGOING_HEALS_ON__S__OFF)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__GRAPH.."|r "..TEXT_SWITCHES_DISPLAY_OF_GRAPH_ON__S__OFF)
--  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__GRAPH.." "..TEXT__SCALE.." X|r "..TEXT_SETS_SCALE_OF_TARGET_GRAPH)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__STATS.."|r "..TEXT_SHOWS_NUMBER_OF_HEALS_AND_CANCELLED_HEALS)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__STATS_RESET.."|r "..TEXT_RESETS_STATISTICS)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__USERS.."|r "..TEXT_SHOWS_A_LIST_OF_PLAYERS_USING_INCOMING_HEALS_IN_YOUR_RAID)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__RESET.."|r "..TEXT_RESETS_ALL_NUMBERS_EXCEPT_STATISTICS)
  echo("|c00"..IH_COLOR_CODE.."/"..TEXT__IH.." "..TEXT__VERSION.."|r  "..TEXT_SHOWS_VERSION)
end

function IH_ShowVersion()
  echo("|c00"..IH_COLOR_CODE..IH_TITLE.."|r "..TEXT__VERSION.." "..IH_VERSION)
  echo(TEXT_AUTHOR..": "..IH_AUTHOR_NAME.." <"..IH_AUTHOR_GUILD.."> "..TEXT__ON.." "..IH_AUTHOR_REALM)
end



registerEvent("VARIABLES_LOADED", IH_OnVariablesLoaded)
registerEvent("CHAT_MSG_SPELL_SELF_BUFF", IH_OnMsgSpellSelfBuff)
registerEvent("SPELLCAST_DELAYED", IH_OnSpellCastDelayed)
registerEvent("PLAYER_TARGET_CHANGED", IH_OnTargetChanged)
registerEvent("SPELLCAST_START", IH_OnSpellCastStart)
registerEvent("SPELLCAST_INTERRUPTED", IH_OnSpellCastStop)
registerEvent("SPELLCAST_STOP", IH_OnSpellCastStop)
registerEvent("SPELLCAST_FAILED", IH_OnSpellCastStop)