--[[
  Various functions for handling strings, tables and printing info to chat window.
  Also has functions to allow Addons that don't use frames
--]]

function echo(str)
  DEFAULT_CHAT_FRAME:AddMessage(str)
end

function print(table, rowname, level) -- like Sea.io.printTable, but output is neater ;)
  if level == nil then
    if type(table) == "table" then level = 1
    else level = 0 end
  end

  local msg = string.rep("  ", level)
  if not rowname then rowname = ""
  else rowname = rowname.." = " end
  
  if table == nil then echo(msg..rowname.."nil") return end
  if type(table) == "table" then
    echo(msg..rowname.."{")
    for k, v in table do
      print(v, k, level + 1)
    end
    echo(msg.."}")
  elseif type(table) == "function" then
    echo(msg..rowname.."{{Function}}")
  elseif type(table) == "userdata" then
    echo(msg..rowname.."{{UserData}}")
  elseif type(table) == "boolean" then
    local value = "true"
    if not table then value = "false" end
    echo(msg..rowname..value)
  else
    echo(msg..rowname..table)
  end
end

SlashCmdList["PRINT"] = function(msg) RunScript("printmsg = ("..msg..")") print(printmsg) end
SLASH_PRINT1 = "/print"
SLASH_PRINT2 = "/z"


ScriptErrors.Show = function()
  echo(ScriptErrors_Message:GetText())
end


function nil2false(val)
  if val == nil then return false else return val end
end
function nil2zero(val)
  if val == nil then return 0 else return val end
end
function nil2string(val)
  if val == nil then return "" else return val end
end

function string2nil(val)
  if val == "" then return nil else return val end
end
function zero2nil(val)
  if val == 0 then return nil else return val end
end
function false2nil(val)
  if val == false then return nil else return val end
end


string.split = function (str, delim)
  local result, i, j = { }, 0, 0
  
  repeat
    i, j = string.find(str, delim, 1, true)
    if i == nil then
      i = string.len(str) + 1
      j = string.len(str)
    end
    table.insert(result, string.sub(str, 1, i - 1))
    str = string.sub(str, j + 1)
  until str == ""

  return result
end

string.join = table.concat





table.append = function (tdest, tsrc)
  for _, value in tsrc do
    table.insert(tdest, value)
  end
end

table.copy = function (tsrc)
  local tdest = { }
  
  for key, value in tsrc do
    if (type(value) == "table") then
      tdest[key] = table.copy(value)
    else
      tdest[key] = value
    end
  end
  
  return tdest
end

function table.getkey(t, func)
  local funcval = func
  if type(func) ~= "function" then func = function(v) return v == funcval end end
  if t then
    for key, value in t do
      if func(value) then return key end
    end
  end
end

function table.keys(t)
  local keys = { }
  for key, _ in t do
    table.insert(keys, key)
  end
  return keys
end








if not RegisteredEvents then RegisteredEvents = { } end

function registerEvent(type, func)
  if not RegisteredEvents[type] then
    RegisteredEvents[type] = { }
    tools:RegisterEvent(type)
  end
  table.insert(RegisteredEvents[type], func)
end

function tools_OnLoad()
  this:RegisterEvent("VARIABLES_LOADED")
end

function tools_OnEvent()
  if RegisteredEvents[event] then
    for _, func in RegisteredEvents[event] do
      func()
    end
  end
  if event == "VARIABLES_LOADED" then
    RegisteredEvents[event] = nil
    TimersActive = true
    tools:UnregisterEvent(event)
  end
end





if not HookedFunctions then HookedFunctions = { } end

function runhooks(funcname, arg)
  local ra1,ra2,ra3,ra4,ra5,ra6,ra7,ra8,ra9,ra10,ra11,ra12,ra13,ra14,ra15,ra16,ra17,ra18,ra19,ra20
  local hookdata = HookedFunctions[funcname]
  for i = 1, table.getn(hookdata) do
    local r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15,r16,r17,r18,r19,r20 = hookdata[i].func(unpack(arg))
    if r1 or r2 or r3 or r4 or r5 or r6 or r7 or r8 or r9 or r10 or r11 or r12 or r13 or r14 or r15 or r16 or r17 or r18 or r19 or r20 then
      ra1,ra2,ra3,ra4,ra5,ra6,ra7,ra8,ra9,ra10,ra11,ra12,ra13,ra14,ra15,ra16,ra17,ra18,ra19,ra20 = r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15,r16,r17,r18,r19,r20
    end
  end
  return ra1,ra2,ra3,ra4,ra5,ra6,ra7,ra8,ra9,ra10,ra11,ra12,ra13,ra14,ra15,ra16,ra17,ra18,ra19,ra20
end

function hook(funcname, func, pos)
  local hookdata = HookedFunctions[funcname]
  if not hookdata then
    hookdata = { }
    HookedFunctions[funcname] = hookdata
    oldfunc = nil
    RunScript("oldfunc = "..funcname)
    table.insert(hookdata, { func = oldfunc, pos = 100 })
    RunScript(funcname.." = function(...) return runhooks('"..funcname.."', arg) end")
  end
  table.insert(hookdata, { func = func, pos = pos })
  table.sort(hookdata, function(t1, t2) return t1.pos < t2.pos end)
end